from aim import Run
from omegaconf import DictConfig

from .device import Device


class Action:
    """
    A class to manage and configure actions based on a configuration and an Aim run session.

    This class initializes an action with specific configurations for task execution,
    sets up device configuration for the action, and ties the action to an Aim run session
    for tracking and logging purposes.

    Attributes:
        cfg (DictConfig): The full configuration object, typically containing all settings.
        action_cfg (DictConfig): A subset of the configuration specific to the action.
        task_cfg (dict): Configuration for the task derived from the action configuration.
        aim_run (Run): An Aim run instance for tracking and logging the execution of the action.
    """

    def __init__(self, cfg: DictConfig, aim_run: Run):
        """
        Initializes the Action object with the necessary configuration and Aim session.

        Args:
            cfg (DictConfig): The configuration object providing settings for various components of the action.
            aim_run (Run): The Aim run session to track and manage execution metrics and logs.

        Sets up the device configuration for the action by calling an external method set_device
        from the Device class with the device setting specified in the action configuration.
        """
        self.cfg = cfg
        self.action_cfg = cfg.action_config
        self.task_cfg = self.action_cfg.get("task")
        self.aim_run = aim_run
        Device.set_device(self.action_cfg.device)



import logging
import os
import sys
from pathlib import Path
import re
import shutil
from abc import ABC, abstractmethod
from typing import List, Optional

from aim import Repo, Run
from hydra.core.hydra_config import HydraConfig
from hydra.core.plugins import Plugins
from omegaconf import OmegaConf

from urartu.utils.hydra_plugin import UrartuPlugin
from urartu.utils.launcher import launch, launch_on_slurm

Plugins.instance().register(UrartuPlugin)

import hydra
from omegaconf import DictConfig


class Command(ABC):
    """Base class for all commands."""
    
    @abstractmethod
    def execute(self) -> None:
        """Execute the command."""
        pass

    @staticmethod
    @abstractmethod
    def get_command_name() -> str:
        """Get the command name used in CLI."""
        pass


class CleanCommand(Command):
    """Command to clean up runs that are not present in the Aim repository."""
    
    def __init__(self, aim_repo_path: str, runs_dir: str):
        self.aim_repo_path = aim_repo_path
        self.runs_dir = runs_dir
        
    @staticmethod
    def get_command_name() -> str:
        return "clean"
        
    def execute(self) -> None:
        # Set up logging
        logging.basicConfig(
            level=logging.INFO,
            format='%(asctime)s - %(levelname)s - %(message)s',
            handlers=[logging.StreamHandler()]
        )

        # Get run hashes from Aim repo
        repo = Repo(self.aim_repo_path)
        run_hashes = repo.list_all_runs()
        logging.info(f"Total number of runs found: {len(run_hashes)}")

        run_dir = Path(self.runs_dir)
        if not run_dir.exists():
            logging.warning(f"Run directory {run_dir} does not exist")
            return

        self._process_action_directories(run_dir, run_hashes)

    def _process_action_directories(self, run_dir: Path, run_hashes: List[str]) -> None:
        """Process each action directory."""
        for action_dir in run_dir.iterdir():
            if not action_dir.is_dir():
                continue

            logging.info(f"Processing action directory: {action_dir.name}")
            self._handle_debug_directory(action_dir)
            self._clean_run_directories(action_dir, run_hashes)

    def _handle_debug_directory(self, action_dir: Path) -> None:
        """Delete debug directory if it exists."""
        debug_dir = action_dir / "debug"
        if debug_dir.exists():
            try:
                shutil.rmtree(debug_dir)
                logging.info(f"Deleted debug directory: {debug_dir}")
            except Exception as e:
                logging.error(f"Failed to delete debug directory: {str(e)}")

    def _clean_run_directories(self, action_dir: Path, run_hashes: List[str]) -> None:
        """Clean run directories that don't have valid hashes."""
        dirs_to_delete = []
        date_pattern = r"\d{4}-\d{2}-\d{2}_\d{2}-\d{2}-\d{2}"

        for dir_path in action_dir.glob("**/"):
            if "_multirun" in str(dir_path):
                self._handle_multirun_directory(dir_path, date_pattern, run_hashes, dirs_to_delete)
            elif re.match(date_pattern, dir_path.name):
                self._handle_regular_directory(dir_path, run_hashes, dirs_to_delete)

        self._delete_marked_directories(action_dir.name, dirs_to_delete)

    def _handle_multirun_directory(self, dir_path: Path, date_pattern: str, 
                                 run_hashes: List[str], dirs_to_delete: List[Path]) -> None:
        """Handle multirun directory checking."""
        if re.match(date_pattern + "_multirun$", dir_path.name):
            all_num_dirs = []
            invalid_num_dirs = []

            for num_dir in dir_path.iterdir():
                if num_dir.is_dir() and num_dir.name.isdigit():
                    all_num_dirs.append(num_dir)
                    if not self._has_valid_hash(num_dir, run_hashes):
                        invalid_num_dirs.append(num_dir)

            if all_num_dirs and len(all_num_dirs) == len(invalid_num_dirs):
                dirs_to_delete.append(dir_path)
                logging.info(f"Multirun directory {dir_path} has no valid runs, marking for deletion")
            elif invalid_num_dirs:
                dirs_to_delete.extend(invalid_num_dirs)
                logging.info(f"Found {len(invalid_num_dirs)} invalid runs in multirun directory {dir_path}")

    def _handle_regular_directory(self, dir_path: Path, run_hashes: List[str], 
                                dirs_to_delete: List[Path]) -> None:
        """Handle regular directory checking."""
        if not self._has_valid_hash(dir_path, run_hashes):
            dirs_to_delete.append(dir_path)
            logging.info(f"Directory {dir_path} has no matching run hash yaml file")

    def _has_valid_hash(self, dir_path: Path, run_hashes: List[str]) -> bool:
        """Check if directory contains a yaml file with a valid hash."""
        yaml_files = list(dir_path.glob("*.yaml"))
        return any(yaml_file.stem in run_hashes for yaml_file in yaml_files)

    def _delete_marked_directories(self, action_name: str, dirs_to_delete: List[Path]) -> None:
        """Delete all marked directories."""
        if dirs_to_delete:
            logging.info(f"Found {len(dirs_to_delete)} directories to delete in action {action_name}")
            for dir_path in dirs_to_delete:
                try:
                    shutil.rmtree(dir_path)
                    logging.info(f"Deleted directory: {dir_path}")
                except Exception as e:
                    logging.error(f"Failed to delete directory {dir_path}: {str(e)}")
        else:
            logging.info(f"No directories to delete in action {action_name}")


class CommandRegistry:
    """Registry for all available commands."""
    
    _commands = {
        CleanCommand.get_command_name(): CleanCommand
    }
    
    @classmethod
    def get_command(cls, command_name: str, **kwargs) -> Optional[Command]:
        """Get a command instance by name."""
        command_class = cls._commands.get(command_name)
        if command_class:
            return command_class(**kwargs)
        return None

    @classmethod
    def register_command(cls, command_class: type) -> None:
        """Register a new command."""
        cls._commands[command_class.get_command_name()] = command_class


def parse_command_args(args: List[str]) -> dict:
    """Parse command arguments in the format key=value.
    
    Args:
        args: List of command line arguments
        
    Returns:
        Dictionary of parsed arguments
    """
    parsed_args = {}
    for arg in args:
        if "=" not in arg:
            continue
        key, value = arg.split("=", 1)
        parsed_args[key] = value
    return parsed_args


def main():
    """Main entry point for the package."""
    if len(sys.argv) > 1 and not any("action_config" in i for i in sys.argv[1:]):
        cli_command = sys.argv[1]
        command_args = parse_command_args(sys.argv[2:])
        
        if cli_command == "clean":
            required_args = {"aim_repo", "runs_dir"}
            missing_args = required_args - set(command_args.keys())
            
            if missing_args:
                raise ValueError(f"Missing required arguments: {', '.join(missing_args)}")
            
            try:
                command = CommandRegistry.get_command(
                    "clean",
                    aim_repo_path=command_args["aim_repo"],
                    runs_dir=command_args["runs_dir"]
                )
                if command:
                    command.execute()
                return
            except Exception as e:
                raise ValueError(f"Failed to execute command: {str(e)}")
        else:
            raise KeyError(f"Unknown command: {cli_command}")

    # If we get here, proceed with normal Hydra execution
    _hydra_main()


@hydra.main(version_base=None, config_path="config", config_name="main")
def _hydra_main(cfg: DictConfig) -> None:
    """Hydra main function for running experiments."""
    hydra_cfg = HydraConfig.get()
    cfg = OmegaConf.create(OmegaConf.to_container(cfg, resolve=True, enum_to_str=True))

    cwd = Path.cwd()

    # Verify current directory is a Python module
    is_module = cwd.joinpath("__init__.py").exists()
    if not is_module:
        logging.error(
            f"The directory '{cwd}' is not recognized as a Python module because it lacks an '__init__.py' file."
            " To resolve this issue, ensure that an '__init__.py' file exists in the directory if it's intended to be a Python module."
        )
        raise FileNotFoundError("Missing '__init__.py' file.")

    # Verify actions directory exists
    are_actions = cwd.joinpath("actions").is_dir()
    if not are_actions:
        logging.error(
            f"The required 'actions' subdirectory does not exist in the directory '{cwd}'."
            " Please ensure that an 'actions' directory is created within the current directory to proceed."
        )
        raise FileNotFoundError("Missing 'actions' directory.")

    # Verify specific action file exists
    action_file_path = cwd.joinpath("actions", f"{cfg.action_name}.py")
    if action_file_path.exists():
        logging.info(f"The action file '{action_file_path}' is located and is ready to be used!")
    else:
        logging.error(
            f"The action file '{action_file_path}' does not exist."
            " Please ensure that the file is correctly named and located in the 'actions' directory."
        )
        raise FileNotFoundError("Missing action file.")

    is_multirun = hydra_cfg.mode.name == "MULTIRUN"

    if is_multirun:
        run_dir = Path(hydra_cfg.runtime.output_dir, hydra_cfg.job.id).parent
    else:
        run_dir = Path(cfg.run_dir)
        if cfg.debug:
            parts = list(Path(run_dir).parts)
            parts.insert(-1, "debug")
            run_dir = Path(*parts)
        os.makedirs(run_dir, exist_ok=True)

    aim_run = None
    if cfg.aim.use_aim:
        aim_run = Run(
            repo=cfg.aim.repo,
            experiment=cfg.action_config.experiment_name,
            log_system_params=cfg.aim.log_system_params,
        )
        aim_run.set("cfg", cfg, strict=False)
        if cfg.debug:
            aim_run.add_tag("debug")
        cfg.aim.hash = aim_run.hash

    if cfg.aim.use_aim:
        log_file = run_dir.joinpath(f"{aim_run.hash}.log")
    else:
        log_file = run_dir.joinpath("output.log")

    root_logger = logging.getLogger()
    root_logger.setLevel(logging.INFO)
    file_handler = logging.FileHandler(log_file)
    stream_handler = logging.StreamHandler()
    formatter = logging.Formatter("%(asctime)s - %(levelname)s - %(message)s")
    file_handler.setFormatter(formatter)
    stream_handler.setFormatter(formatter)
    root_logger.handlers = [file_handler, stream_handler]

    class TeeHandler:
        def __init__(self, filename, stream):
            self.terminal = stream
            self.log = open(filename, 'a')

        def write(self, message):
            self.terminal.write(message)
            self.log.write(message)
            self.flush()

        def flush(self):
            self.terminal.flush()
            self.log.flush()

        def fileno(self):
            return self.terminal.fileno()

        def isatty(self):
            return self.terminal.isatty()

        def close(self):
            self.log.close()

    sys.stdout = TeeHandler(log_file, sys.stdout)
    sys.stderr = TeeHandler(log_file, sys.stderr)

    cfg.run_dir = str(run_dir)
    if cfg.aim.use_aim:
        with open(run_dir.joinpath(f"{aim_run.hash}.yaml"), "w") as f:
            OmegaConf.save(config=cfg, f=f)
    else:
        with open(run_dir.joinpath("cfg.yaml"), "w") as f:
            OmegaConf.save(config=cfg, f=f)

    try:
        if cfg.slurm.use_slurm:
            try:
                import submitit  # NOQA
            except ImportError:
                raise ImportError("Please 'pip install submitit' to schedule jobs on SLURM")

            try:
                launch_on_slurm(
                    module=cwd,
                    action_name=cfg.action_name,
                    cfg=cfg,
                    aim_run=aim_run,
                )
            except submitit.core.utils.FailedJobError as e:
                logging.error(f"Slurm job failed: {e}")
                raise
            except submitit.core.utils.FailedSubmissionError as e:
                logging.error(f"Failed to submit job to SLURM: {e}")
                raise
            except RuntimeError as e:
                if "Could not detect 'srun'" in str(e):
                    logging.error("Not running on a SLURM cluster or 'srun' command not available")
                else:
                    logging.error(f"Runtime error during SLURM job execution: {e}")
                raise
        else:
            try:
                launch(
                    module=cwd,
                    action_name=cfg.action_name,
                    cfg=cfg,
                    aim_run=aim_run,
                )
            except ImportError as e:
                logging.error(f"Failed to import required module for local execution: {e}")
                raise
            except Exception as e:
                logging.error(f"Error during local job execution: {e}")
                raise
    except Exception as e:
        logging.error(f"Unexpected error: {e}")
        raise
    finally:
        if aim_run:
            aim_run.close()


if __name__ == "__main__":
    main()



from typing import Tuple, Union

import tiktoken
from langchain.schema import HumanMessage
from langchain_openai import AzureChatOpenAI
from transformers import AutoModelForCausalLM

from urartu.common.model import Model


class ModelOpenAI(Model):
    """
    A class for interacting with an OpenAI model deployed through Azure, facilitating
    the initialization, configuration, and use of the model for text generation.

    Attributes:
        cfg: A configuration object containing model deployment details and API keys for Azure.

    Methods:
        model: Returns an instance of the model ready for interaction.
        generate: Generates responses based on input prompts.
        _get_num_tokens: Calculates the number of tokens in a given string based on the specified model encoding.
    """

    def __init__(self, cfg) -> None:
        """
        Initializes the ModelOpenAI class by setting up the base model configuration.

        Args:
            cfg: Configuration object containing parameters such as the name of the Azure deployment,
                 API types, API versions, endpoint, and API keys.
        """
        super().__init__(cfg)

    @property
    def model(self) -> AutoModelForCausalLM:
        """
        Retrieves or creates an instance of AzureChatOpenAI as a causal language model,
        configured as specified in the class configuration.

        Returns:
            An instance of AzureChatOpenAI configured and ready to interact.
        """
        if self._model is None:
            self._model = AzureChatOpenAI(
                deployment_name=self.cfg.name,
                openai_api_type=self.cfg.openai_api_type,
                openai_api_version=self.cfg.openai_api_version,
                azure_endpoint=self.cfg.azure_openai_endpoint,
                openai_api_key=self.cfg.azure_openai_api_key,
            )
        return self._model

    def generate(self, prompt: Union[str, Tuple[str, str]], generate_cfg):
        """
        Generates a response from the model based on the provided prompt and generation configurations.

        Args:
            prompt: A string or a tuple containing the prompt(s) to be sent to the model.
            generate_cfg: Configuration parameters specifying how the model should generate the response.

        Returns:
            A string containing the generated response from the model.
        """
        output = self.model(HumanMessage(content=prompt))

        return output

    def _get_num_tokens(self, string: str, encoding_name: str = "gpt-3.5-turbo") -> int:
        """
        Computes the number of tokens that the specified string will be broken into, using a particular model encoding.

        Args:
            string: The text string to be tokenized.
            encoding_name: The name of the encoding model to be used (default is "gpt-3.5-turbo").

        Returns:
            An integer representing the number of tokens.
        """
        encoding = tiktoken.encoding_for_model(encoding_name)
        num_tokens = len(encoding.encode(string))
        return num_tokens



import torch
from transformers import AutoModelForCausalLM, AutoTokenizer

from urartu.common.device import Device
from urartu.common.model import Model
from urartu.utils.dtype import eval_dtype


class ModelCausalLanguage(Model):
    """
    A class for working with a causal language model (CLM) that handles model initialization,
    configuration, and text generation using the model and tokenizer from the Hugging Face
    Transformers library.

    Attributes:
        cfg: An object containing configuration settings such as model name, cache directory,
             API token, and other model-specific settings.

    Methods:
        model: Returns a causal language model with preloaded configurations.
        tokenizer: Returns a tokenizer associated with the causal language model.
        generate: Generates text based on a given prompt and optional generation settings.
    """

    def __init__(self, cfg) -> None:
        """
        Initializes the ModelCausalLanguage class by setting up the base model configuration.

        Args:
            cfg: Configuration object containing necessary parameters like model name,
                 cache directory, device mapping, dtype evaluation, and optional API token.
        """
        super().__init__(cfg)
        self._tokenizer = None

    @property
    def model(self) -> AutoModelForCausalLM:
        """
        Retrieves or instantiates the causal language model specified in the configuration.
        The model is set to evaluation mode with gradients disabled.

        Returns:
            An instance of AutoModelForCausalLM ready for inference.
        """
        if self._model is None:
            self._model = AutoModelForCausalLM.from_pretrained(
                self.cfg.name,
                cache_dir=self.cfg.cache_dir,
                device_map=Device.get_device(),
                torch_dtype=eval_dtype(self.cfg.dtype),
                token=self.cfg.api_token,
                trust_remote_code=self.cfg.get("trust_remote_code"),
            )
            for param in self._model.parameters():
                param.requires_grad = False
            self._model.eval()
        return self._model

    @property
    def tokenizer(self):
        """
        Retrieves or instantiates the tokenizer associated with the causal language model,
        setting the pad token to be the same as the eos token if not already set.

        Returns:
            An instance of AutoTokenizer.
        """
        if self._tokenizer is None:
            self._tokenizer = AutoTokenizer.from_pretrained(self.cfg.name)
            self._tokenizer.pad_token = self._tokenizer.eos_token
        return self._tokenizer

    def generate(self, prompt: str, generate_cfg=None):
        """
        Generates text based on a provided prompt and optional generation settings.

        Args:
            prompt: A string containing the initial text to prompt the model with.
            generate_cfg: An optional dictionary containing generation parameters such as
                          maximum length, number of beams, etc. Defaults to configurations
                          specified in the model's config.

        Returns:
            A string of generated text. If 'output_scores' is included in the generation
            configuration, it returns a tuple of the generated text and the corresponding
            generation scores.
        """
        if not generate_cfg:
            generate_cfg = self.cfg.get("generate")
        self.model.eval()

        prompt_tokenized = self.tokenizer(prompt, return_tensors="pt", padding=True, truncation=True)
        prompt_tensor = prompt_tokenized["input_ids"].to(self.model.device)
        attention_mask = prompt_tokenized["attention_mask"].to(self.model.device)
        with torch.no_grad():
            output_tokenized = self.model.generate(input_ids=prompt_tensor, attention_mask=attention_mask, **generate_cfg)
        if "output_scores" in generate_cfg:
            output = self.tokenizer.decode(output_tokenized["sequences"][0], skip_special_tokens=True)
            return output, output_tokenized["scores"]
        else:
            return self.tokenizer.decode(output_tokenized[0], skip_special_tokens=True)
